"""Module to store constants used in the application."""

PRIMARY_COLOR = '#68B7FF'
SECONDARY_COLOR = '#363F4B'
SECONDARY_COLOR_LIGHT = '#ABA7A7'
TEXT_COLOR = '#FFFFFF'

INFO_COLOR = '#2196F3'
DANGER_COLOR = '#FF3F51'
WARNING_COLOR = '#FFC107'
SUCCESS_COLOR = '#03F7AE'
